(* konvertieren von Postscript nach PDF mit Ghostscript

    2021 Dr. J. Rathlev, D-24222 Schwentinental (info(a)rathlev-home.de)

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   J. Rathlev, Mrz 2005
   Vers. 2.1 (Delphi XE2), Aug. 2013
   Vers. 3.0 (Delphi 10), Mai 2021
   *)

unit Ps2PdfMain;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes,
  Vcl.Graphics, Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls, Vcl.Buttons,
  HListBox, Vcl.ComCtrls, Vcl.ExtCtrls, Prgrss;

const
  Prog = 'PS to PDF';
  Vers = ' Vers. 3.1';
  CopRgt = ' 2005 - 2021 Dr. J. Rathlev, 24222 Schwentinental';
  EmailAdr = 'kontakt(a)rathlev-home.de';

  defPipeSize = 16*1024;

// Schlssel zum Kodieren der Passwrter in der Ini-Datei
  PwdKey = '6B93AFCD27B0D081F5E7E2815D25C4C3416026737D163AF1'; //54AFqnU2
  MaxUserPwdLength = 64;

  pdfSettings : array [0..4] of string =('screen','ebook','printer','prepress','default');
  pmRead     = 1;
  pmPrint    = 2;
  pmLowPrint = 4;
  pmExtract  = 8;
  pmEdit     = 16;
  pmFill     = 32;

  pbitDefault  = -4;
  pbitPrint    = -5;         // Bit 3
  pbitModify   = -9;         // Bit 4
  pbitCopy     = -17;        // Bit 5
  pbitAddFill  = -33;        // Bit 6
  pbitFill     = -257;       // Bit 9
  pbitExtract  = -513;       // Bit 10
  pbitAssemble = -1025;      // Bit 11
  pbitLowPrint = -2049;      // Bit 12
  
type
  TfrmPs2Pdf = class(TForm)
    PageControl: TPageControl;
    TabSheet1: TTabSheet;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    btnPsDir: TSpeedButton;
    btnPsFiles: TSpeedButton;
    edtPdfDir: THistoryCombo;
    edtPsDir: THistoryCombo;
    edtPsFiles: THistoryCombo;
    OpenDialog: TOpenDialog;
    TabSheet2: TTabSheet;
    DestSelectBtn: TSpeedButton;
    btnQuit: TBitBtn;
    btnInfo: TSpeedButton;
    cbEncrypt: TCheckBox;
    gbxEncrypt: TGroupBox;
    Label23: TLabel;
    Label25: TLabel;
    Label26: TLabel;
    edtPassword: TEdit;
    edtPwdAgain: TEdit;
    rdb40: TRadioButton;
    rdb128: TRadioButton;
    gbxGhostscript: TGroupBox;
    Label24: TLabel;
    btnGsDir: TSpeedButton;
    edtGsDir: TEdit;
    gpbPermissions: TGroupBox;
    cbNoLimit: TCheckBox;
    cbLowPrint: TCheckBox;
    cbNoPrint: TCheckBox;
    cbNoExtract: TCheckBox;
    cbNoEdit: TCheckBox;
    cbNoFill: TCheckBox;
    rgpSettings: TRadioGroup;
    BarProgress: TBarProgress;
    lblStatus: TLabel;
    btnConvert: TBitBtn;
    procedure btnPsDirClick(Sender: TObject);
    procedure DestSelectBtnClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure btnQuitClick(Sender: TObject);
    procedure btnPsFilesClick(Sender: TObject);
    procedure edtPwdAgainExit(Sender: TObject);
    procedure btnGsDirClick(Sender: TObject);
    procedure btnConvertClick(Sender: TObject);
    procedure cbNoLimitClick(Sender: TObject);
    procedure cbLimitsClick(Sender: TObject);
    procedure cbNoPrintClick(Sender: TObject);
    procedure cbLowPrintClick(Sender: TObject);
    procedure btnInfoClick(Sender: TObject);
  private
    { Private-Deklarationen }
    DefPath,Ininame : string;
    function ExecuteProcess (Cmd,Path : string) : boolean;
    function PsToPdf (GsPath,PsName,PdfPath,Setting,Key,Pwd : string;
                      Perms : integer) : boolean;
  public
    { Public-Deklarationen }
  end;

var
  frmPs2Pdf: TfrmPs2Pdf;

implementation

{$R *.dfm}

uses System.IniFiles, Crypt, WinUtils, WinApiUtils, WinShell, Winapi.ShlObj,
  System.StrUtils, StringUtils, FileUtils, StatWind, Show, ShellDirDlg;

const
  IniExt = 'ini';
  PsExt  = '.ps';
  PdfExt = '.pdf';

  (* INI-Sektionen *)
  CfGSekt = 'Config';
  SourceSekt = 'Source';
  DestSekt = 'Dest';
  FileSekt = 'Files';

  (* INI-Variablen *)
  iniGsDir = 'GsDir';
  iniKey   = 'Keylength';
  iniPwd   = 'Password';
  iniSet   = 'Settings';
  iniPerm  = 'Permissions';

procedure TfrmPs2Pdf.FormCreate(Sender: TObject);
var
  IniFile  : TIniFile;
  n        : integer;
begin
  DefPath:=GetDesktopFolder(CSIDL_PERSONAL);
  IniName:=Erweiter(GetDesktopFolder(CSIDL_APPDATA),
    ExtractFileName(DelExt(Application.ExeName)),IniExt);
  IniFile:=TIniFile.Create(IniName);
  with IniFile do begin
    edtGsDir.Text:=ReadString(CfgSekt,iniGsDir,'C:\gs');
    n:=ReadInteger(CfgSekt,iniPerm,0);
    if n=0 then cbNoLimit.Checked:=true
    else begin
      cbEncrypt.Checked:=n and pmRead <>0;
      cbNoPrint.Checked:=n and pmPrint <>0;
      cbLowPrint.Checked:=n and pmLowPrint <>0;
      cbNoExtract.Checked:=n and pmExtract <>0;
      cbNoEdit.Checked:=n and pmEdit <>0;
      cbNoFill.Checked:=n and pmFill <>0;
      end;
    n:=ReadInteger(CfgSekt,iniKey,1);
    if n=0 then rdb40.Checked:=true else rdb128.Checked:=true;
    edtPassword.Text:=DecryptPwdString(DecodeString(PwdKey),ReadString(CfgSekt,iniPwd,''),teAscii85);
    edtPwdAgain.Text:=edtPassword.Text;
    rgpSettings.ItemIndex:=ReadInteger(CfgSekt,iniSet,0);
    Free;
    end;
  edtPsDir.LoadFromIni (IniName,SourceSekt);
  edtPsDir.ItemIndex:=0;
  edtPsFiles.LoadFromIni (IniName,DestSekt);
  edtPsFiles.ItemIndex:=0;
  edtPdfDir.LoadFromIni (IniName,FileSekt);
  edtPdfDir.ItemIndex:=0;
  lblStatus.Caption:='';
  end;

procedure TfrmPs2Pdf.FormDestroy(Sender: TObject);
var
  IniFile  : TIniFile;
  n        : integer;
begin
  IniFile:=TIniFile.Create(IniName);
  with IniFile do begin
    WriteString(CfgSekt,iniGsDir,edtGsDir.Text);
    n:=0;
    if not cbNoLimit.Checked then begin
      if cbEncrypt.Checked then n:=n or pmRead;
      if cbNoPrint.Checked then n:=n or pmPrint;
      if cbLowPrint.Checked then n:=n or pmLowPrint;
      if cbNoExtract.Checked then n:=n or pmExtract;
      if cbNoEdit.Checked then n:=n or pmEdit;
      if cbNoFill.Checked then n:=n or pmFill;
      end;
    WriteInteger(CfgSekt,iniPerm,n);
    if rdb40.Checked then n:=0 else n:=1;
    WriteInteger(CfgSekt,iniKey,n);
    WriteString(CfgSekt,iniPwd,EncryptPwdString(DecodeString(PwdKey),edtPassword.Text,MaxUserPwdLength,teAscii85));
    writeInteger(CfgSekt,iniSet,rgpSettings.ItemIndex);
    Free;
    end;
  end;

procedure TfrmPs2Pdf.btnPsDirClick(Sender: TObject);
var
  s : string;
begin
  s:=edtPsDir.Text;
  if length(s)=0 then s:=DefPath;
  if ShellDirDialog.Execute ('Verzeichnis mit PS-Dateien',true,true,true,'',s) then begin
    with edtPsDir do begin
      Text:=s; AddItem(s);
      end;
    with edtPdfDir do if length(Text)=0 then begin
      Text:=s; AddItem(s);
      end;
    with edtPsFiles do if length(Text)=0 then begin
      Text:='*'+PsExt; AddItem(Text);
      end;
    end;
  end;

procedure TfrmPs2Pdf.DestSelectBtnClick(Sender: TObject);
var
  s : string;
begin
  s:=edtPdfDir.Text;
  if length(s)=0 then s:=edtPsDir.Text;
  if length(s)=0 then s:=DefPath;
  if ShellDirDialog.Execute ('Verzeichnis fr Pdf-Dateien',true,true,true,edtPsDir.Text,s) then begin
    with edtPdfDir do begin
      Text:=s; AddItem(s);
      end;
    end;
  end;

procedure TfrmPs2Pdf.btnQuitClick(Sender: TObject);
begin
  Close;
  end;

procedure TfrmPs2Pdf.btnPsFilesClick(Sender: TObject);
var
  i : integer;
  s : string;
begin
  with OpenDialog do begin
    InitialDir:=edtPsDir.Text;
    Options:=[ofHideReadOnly,ofAllowMultiSelect,ofPathMustExist,ofFileMustExist,ofEnableSizing];
    Filename:='';
    Filter:='PS-Dateien|*'+PsExt+'|alle|*.*';
    Title:='Postscript-Dateien auswhlen';
    if Execute then begin
      s:=ExtractFilePath(Files.Strings[0]);
      with edtPsDir do if s<>Text then begin
        Text:=s; AddItem(s);
        end;
      with Files do begin
        s:='';
        for i:=0 to Count-1 do s:=s+ExtractFilename(Strings[i])+';';
        end;
      delete (s,length(s),1);
      with edtPsFiles do begin
        Text:=s; AddItem(s);
        end;
      end;
    end;
  end;

procedure TfrmPs2Pdf.edtPwdAgainExit(Sender: TObject);
begin
  if edtPassword.Text<>edtPwdAgain.Text then begin
    ErrorDialog ('Die Eingaben fr das Passwort stimmen nicht berein!');
    with edtPassword do begin
      Text:=''; SetFocus;
      end;
    edtPwdAgain.Text:='';
    end;
  end;

procedure TfrmPs2Pdf.btnGsDirClick(Sender: TObject);
var
  s : string;
begin
  s:=edtGsDir.Text;
  if ShellDirDialog.Execute ('Ghostscript-Verzeichnis:',true,true,true,'',s) then edtGsDir.Text:=s;
  end;

{ ------------------------------------------------------------------- }
// Prozess mit Befehlszeile "cmd" starten
function TfrmPs2Pdf.ExecuteProcess (Cmd,Path : string) : boolean;
const
  BUFSIZE = 1024;
var
  si        : TStartupInfo;
  pi        : TProcessInformation;
  saAttr    : TSecurityAttributes;
  hChildStdoutRd,
  hChildStdoutWr  : THandle;
  chBuf       : array [0..BUFSIZE] of AnsiChar;
  dwRead,ec   : DWord;
  s           : string;
  sa          : RawByteString;
  j           : integer;
begin
  Result:=false;
  Screen.Cursor:=crHourglass;
  with saAttr do begin
    nLength:=sizeof(SECURITY_ATTRIBUTES);
    bInheritHandle:=TRUE;
    lpSecurityDescriptor:=nil;
    end;
// Create a pipe for the child process's STDOUT.
  CreatePipe(hChildStdoutRd,hChildStdoutWr,@saAttr,defPipeSize);
  SetHandleInformation(hChildStdoutRd, HANDLE_FLAG_INHERIT,0);
// Create process to start compiler
  FillChar(si, SizeOf(TStartupInfo), 0);
  with si do begin
    cb := Sizeof(TStartupInfo);
    dwFlags := STARTF_USESTDHANDLES or STARTF_USESHOWWINDOW;
    wShowWindow:=SW_HIDE;
    hStdOutput:=hChildStdoutWr;
    end;
  try
     if CreateProcess(nil,                // Anwendungsname
                   pchar(cmd),
                   nil,                // Security
                   nil,                // Security
                   true,               // use InheritHandles
                   NORMAL_PRIORITY_CLASS, // Prioritt
                   nil,                   // Environment
                   pchar(Path),     // Verzeichnis
                   si,pi) then begin
      while WaitForSingleObject(pi.hProcess,10)=WAIT_TIMEOUT do Application.ProcessMessages;
      GetExitCodeProcess(pi.hProcess,ec);
      CloseHandle(pi.hProcess);

      frmShow.memShow.Clear;
// Close the write end of the pipe before reading from the
// read end of the pipe.
      if CloseHandle(hChildStdoutWr) then begin
  // Read output from the child process, and write to parent's STDOUT.
        while ReadFile(hChildStdoutRd,chBuf[0],BUFSIZE,dwRead,nil)
              and (dwRead=BUFSIZE) do begin
          sa:=sa+chBuf;
          end;
        if dwRead>0 then begin
          chBuf[dwread]:=#0;
          sa:=sa+chBuf;
          end;
        s:=RawByteToUnicode(sa,850);
        s:=ReplaceStr(s,CrLf,Lf); // Convert Unix style output
        s:=ReplaceStr(s,Lf,CrLf);
        frmShow.memShow.SetSelTextBuf(PChar(s));
        end;
      CloseHandle(hChildStdoutRd);
      // DOS-Ausgabe anzeigen
      with frmShow,memShow do begin
        SelLength:=0;
        with Lines do for j:=Count-1 downto 0 do
          if AnsiLeftStr(Strings[j],2)='%%' then Delete(j);
        if Lines.Count>0 then ShowModal
        else Result:=true;
//        Perform(WM_VSCROLL,SB_BOTTOM,0);
        end;
      end
    else ErrorDialog('Fehler beim Starten von Ghostscript!'+sLineBreak+SysErrorMessage(GetLastError));
  finally
    Screen.Cursor:=crDefault;
    end;
  end;

{ ------------------------------------------------------------------- }
// von PS nach PDF konvertieren
function TfrmPs2Pdf.PsToPdf (GsPath,PsName,PdfPath,Setting,Key,Pwd : string;
                             Perms : integer) : boolean;
const
  GsFonts = '\fonts';
  GsLib   = '\lib';
  GsName  = '\bin\gswin32c.exe';

  GsOptions1 = ' -q -dBATCH -dNOPAUSE -P- ';
  GsOptions2 = ' -sDEVICE#pdfwrite -sPAPERSIZE#a4 -dCompatibilityLevel#1.4';
  GsInclude  = '-I';
  GsEncrypt  = ' -dEncryptionR#3 -dKeyLength#';
  GsOwnerPwd = ' -sOwnerPassword#';
  GsUserPwd  = ' -sUserPassword#';
  GsPermiss  = ' -dPermissions#';  // nur drucken ist erlaubt
  GsSetting  = ' -dPDFSETTINGS#/';
  GsOutput   = '-sOutputFile#';
  BUFSIZE = 1024;
var
  cmd,pf,gf : string;
  j         : integer;
begin
  j:=LastDelimiter('\',GsPath);
  gf:=MakeQuotedStr(copy(GsPath,1,j-1)+GsFonts,[#32]); // Font path
  pf:=ChangeFileExt(PsName,PdfExt);
  pf:=IncludeTrailingPathDelimiter(PdfPath)+ExtractFileName(pf);
  cmd:=MakeQuotedStr(GsPath+GsName,[#32])+GsOptions1+
    MakeQuotedStr(GsInclude+GsPath+GsLib,[#32])+GsOptions2;
  cmd:=cmd+GsSetting+Setting;
  if not cbNoLimit.Checked and (length(Pwd)>0) then begin
    if cbEncrypt.Checked then cmd:=cmd+GsEncrypt+Key+GsOwnerPwd+Pwd+GsUserPwd+Pwd
    else cmd:=cmd+GsEncrypt+Key+GsOwnerPwd+Pwd;
    cmd:=cmd+GsPermiss+IntToStr(Perms);
    end;
  cmd:=cmd+' '+MakeQuotedStr(GsOutput+pf,[#32])+' '+MakeQuotedStr(PsName,[#32]);
  Result:=ExecuteProcess (cmd,GsPath);
  if not Result then ErrorDialog('Datei: '+PsName+#13' konnte nicht in PDF konvertiert werden!');
  end;

procedure TfrmPs2Pdf.btnConvertClick(Sender: TObject);
var
  FListe     : TStringList;
  n,i,pm     : integer;
  fn,dg,
  st,ky,
  s,t        : string;

  procedure ScanDir (Dir,Mask : string);
  var
    DirInfo    : TSearchRec;
    Findresult : integer;
  begin
(*    if chbSubDir.Checked then begin
      FindResult:=FindFirst (Erweiter(Dir,'*','*'),faAnyFile,DirInfo);
      while FindResult=0 do with DirInfo do begin
        if NotSpecialDir(Name) then begin
          if((Attr and faDirectory)<>0) then ScanDir (Erweiter(Dir,Name,''),Mask);
          end;
        FindResult:=FindNext (DirInfo);
        end;
      FindClose(DirInfo);
      end; *)
    FindResult:=FindFirst (Erweiter(Dir,Mask,''),faArchive+faReadOnly+faHidden+faSysfile,DirInfo);
    while FindResult=0 do with DirInfo do begin
      if NotSpecialDir(Name) then FListe.Add(Erweiter(Dir,Name,''));
      FindResult:=FindNext (DirInfo);
      end;
    FindClose(DirInfo);
    end;


begin
  dg:=ExcludeTrailingPathDelimiter(edtGsDir.Text);
  if DirectoryExists(dg) then begin
    t:=edtPsFiles.Text;
    if length(t)>0 then begin
      st:=pdfSettings[rgpSettings.ItemIndex];
      if rdb40.Checked then ky:='40' else ky:='128';
      pm:=pbitDefault;
      if cbNoPrint.Checked then pm:=pm and pbitPrint
      else if cbLowPrint.Checked then pm:=pm and pbitLowPrint;
      if cbNoExtract.Checked then pm:=pm and pbitCopy and pbitExtract;
      if cbNoEdit.Checked then pm:=pm and pbitModify and pbitAssemble;
      if cbNoFill.Checked then pm:=pm and pbitAddFill and pbitFill;
      FListe:=TStringList.Create;
      n:=0;
//      StatusWindow.ShowStatus(CursorPos,'Konvertiere in PDF','',false,5);
      BarProgress.Status:=psOn;
      repeat
        s:=ReadNxtStr (t,';');
        FListe.Clear;
        ScanDir (edtPsDir.Text,s);
        with FListe do if Count>0 then for i:=0 to Count-1 do begin
          fn:=StripPath(Strings[i],45);
          lblStatus.Caption:='Datei: '+StripPath(fn,50);
//          StatusWindow.StatusLabel.Caption:='Datei: '+fn;
          if PsToPDF (dg,Strings[i],edtPdfDir.Text,st,ky,edtPassword.Text,pm) then inc(n);
          end;
        until length(t)=0;
      FListe.Free;
      with BarProgress do begin
        Status:=psOff; Reset;
        end;
//      StatusWindow.Close;
      if n>0 then begin
        if n>1 then //ShowInfo('Datei:'+fn+' wurde konvertiert!')
//        else ShowInfo('Im Verzeichnis: '+edtPsDir.Text+#13'wurden '+IntToStr(n)+' Dateien konvertiert!');
        lblStatus.Caption:=IntToStr(n)+' Dateien konvertiert!';
        end
      else // ShowInfo('Keine passenden Dateien gefunden!');
        lblStatus.Caption:='Keine passenden Dateien gefunden!';
      end
    end
  else ErrorDialog('Ghostscript nicht gefunden!');
  end;

procedure TfrmPs2Pdf.cbNoLimitClick(Sender: TObject);
begin
  if cbNoLimit.Checked then begin
    cbEncrypt.Checked:=false;
    cbLowPrint.Checked:=false;
    cbNoPrint.Checked:=false;
    cbNoExtract.Checked:=false;
    cbNoEdit.Checked:=false;
    cbNoFill.Checked:=false;
    end;
  cbLimitsClick(Sender);
  end;

procedure TfrmPs2Pdf.cbLimitsClick(Sender: TObject);
begin
  cbNoLimit.Checked:=not (cbEncrypt.Checked or cbLowPrint.Checked or cbNoPrint.Checked or
      cbNoExtract.Checked or cbNoEdit.Checked or cbNoFill.Checked);
  end;

procedure TfrmPs2Pdf.cbNoPrintClick(Sender: TObject);
begin
  if cbNoPrint.Checked then cbLowPrint.Checked:=false;
  cbLimitsClick(Sender);
  end;

procedure TfrmPs2Pdf.cbLowPrintClick(Sender: TObject);
begin
  if cbLowPrint.Checked then cbNoPrint.Checked:=false;
  cbLimitsClick(Sender);
  end;

procedure TfrmPs2Pdf.btnInfoClick(Sender: TObject);
begin
  InfoDialog(GetFileVersionName(Application.Exename,Prog,Vers)+sLineBreak+CopRgt
           +sLineBreak+'E-Mail: '+EmailAdr);
  end;

end.
